-- the app size
local myappwidth, myappheight = application:getContentWidth(), application:getContentHeight()

-- react physics 3D
r3d = require "reactphysics3d"
local world = r3d.World.new(0, -9.8, 0)
-- a fullscreen 3D viewport
local view = D3.View.new(myappwidth, myappheight, 45, 0.1, 2048) -- function D3View:init(width, height, fov, near, far)

-- some vars
world.objects = {} -- objects list
local x = 1 -- objects index

-- a STATIC_BODY plane factory
local function plane_factory(params)
	-- params
	local params = params or {}
	local xposx = params.posx or 0
	local xposy = params.posy or 0
	local xposz = params.posz or 0
	local xsizex = params.sizex or 1
	local xsizey = params.sizey or 1
	local xsizez = params.sizez or 1
	local xrotx = params.rotx or 0
	local xroty = params.roty or 0
	local xrotz = params.rotz or 0
	local xtexpath = params.texpath or "3d/texs/box.png"
	local xtexw = params.texw or 8 * 1024
	local xtexh = params.texh or 8 * 1024
	-- plane mesh
	local planemesh = D3.Mesh.new()
	planemesh:setVertexArray{ -xsizex,xsizey,-xsizez,	xsizex,xsizey,-xsizez,	xsizex,xsizey,xsizez,	-xsizex,xsizey,xsizez }
	-- texture coordinates -- make it big (3200, 3200)
	planemesh:setTextureCoordinateArray	{	0, 0,	xtexw, 0,	xtexw, xtexh,	0, xtexh	}
	-- normal array
	planemesh:setGenericArray( 3, Shader.DFLOAT, 3, 4, { 0,1,0, 0,1,0, 0,1,0, 0,1,0 } )
	planemesh:setIndexArray { 1,2,3,	1,3,4 }
	planemesh:setTexture( Texture.new( xtexpath, true, { wrap = TextureBase.REPEAT } ) )
	-- lighting, shadows, and texture. setTexture doesn't automatically do it
	planemesh:updateMode( D3.Mesh.MODE_LIGHTING | D3.Mesh.MODE_SHADOW | D3.Mesh.MODE_TEXTURE )
	-- put it in a viewport so we can matrix it
	local planeview = Viewport.new()
	planeview:setContent(planemesh)
	-- transform
	matrix = planeview:getMatrix()
	matrix:setPosition(xposx, xposy, xposz)
	matrix:setRotationX(xrotx)
	matrix:setRotationY(xroty)
	matrix:setRotationZ(xrotz)
	planeview:setMatrix(matrix)
	-- body
	planeview.body = world:createBody(planeview:getMatrix())
	planeview.body:setType(r3d.Body.STATIC_BODY)
	planemesh.shape = r3d.BoxShape.new(xsizex, xsizey, xsizez)
	planeview.body:createFixture(planemesh.shape, nil, 0) -- shape, transform, weight
	-- add to objects list
	world.objects[x] = planeview
	x += 1
	
	return planeview, planeview.body
end

-- a cube factory
local function cube_factory(params)
	-- params
	local params = params or {}
	xcontinuous = params.continuous or false
	xposx = params.posx or 0
	xposy = params.posy or 0
	xposz = params.posz or 0
	xsizex = params.sizex or 1
	xsizey = params.sizey or 1
	xsizez = params.sizez or 1
	xrotx = params.rotx or 0
	xroty = params.roty or 0
	xrotz = params.rotz or 0
	xtexpath = params.texpath or "3d/texs/box.png"
	-- the box mesh
	local cube = D3.Cube.new(xsizex, xsizey, xsizez)
	cube:mapTexture(Texture.new(xtexpath, true))
	cube:updateMode( D3.Mesh.MODE_LIGHTING | D3.Mesh.MODE_SHADOW )
	-- put it in a viewport so we can matrix it
	local cubeview = Viewport.new()
	cubeview:setContent(cube)
	-- transform
	matrix = cubeview:getMatrix()
	matrix:setPosition(xposx, xposy, xposz)
	matrix:setRotationX(xrotx)
	matrix:setRotationY(xroty)
	matrix:setRotationZ(xrotz)
	cubeview:setMatrix(matrix)
	-- body
	cubeview.body = world:createBody(cubeview:getMatrix())
	cube.shape = r3d.BoxShape.new(xsizex, xsizey, xsizez)
	cubeview.body:createFixture(cube.shape, nil, 1)
	-- add to list
	if xcontinuous then
		world.boxes[cubeview] = cubeview.body
	else
		world.objects[x] = cubeview
		x += 1
	end

	return cubeview, cubeview.body
end

-- LET'S GO!
-- some ground planes
local ground = plane_factory( { sizex=64, sizez=64, texpath="3d/texs/groundx01.png" } )
-- and 2 cubes
local cube01, cube01body = cube_factory( { posx = 0, posy = 12, posz = 2, sizex = 1.5 } )
local cube02, cube02body = cube_factory( { posx = 4, posy = 12, posz = 2, sizex = 1 } )

-- trying joints
--GIDEROS JOINT TEST 1: r3d.World:createBallAndSocketJoint(bodyA,bodyB,anchorX,anchorY,anchorZ,params)
--world:createBallAndSocketJoint(cube01body, cube02body, 0, 0, 0) -- THIS LINE CRASHES THE APP

--GIDEROS JOINT TEST 2: r3d.World:createFixedJoint(bodyA,bodyB,anchorX,anchorY,anchorZ,params)
--world:createFixedJoint(cube01body, cube02body, 0, 0, 0, {"isCollisionEnabled"=false}) -- THIS LINE CRASHES THE APP
--world:createFixedJoint(cube01body, cube02body, 0.5, 0, 0) -- THIS LINE CRASHES THE APP

-- the scene
local scene = view:getScene() -- a 3D View sprite
for o = 1, #world.objects do -- adds the objects
	scene:addChild(world.objects[o])
end
view:lookAt(0, 10, -20, 0, 5, 0) -- initial camera lookat
stage:addChild(view)

-- the light
Lighting.setLight(15, 30, 0, 0.3)
Lighting.setLightTarget(0, 0, 0, 40, 120)

-- the game loop
stage:addEventListener(Event.ENTER_FRAME, function(e)
	-- lighting
	Lighting.computeShadows(scene)
	world:step(e.deltaTime)
	local m1 = cube01body:getTransform()
	cube01:setMatrix(m1)
	local m2 = cube02body:getTransform()
	cube02:setMatrix(m2)
end)
